#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Default values for options, see man page for details. */
#define DEF_B	61	/* For A4, use 36 for letter size paper. */
#define DEF_G	75
#define DEF_H	720
#define DEF_L	28	/* For A4, use 18 for letter size paper. */
#define DEF_P	50
#define DEF_R	0	/* Uses DEF_P if 0 */
#define DEF_S	18
#define DEF_T	10
#define DEF_V	0	/* Uses DEF_S if 0 */
#define DEF_W	540
#define DEF_X	1
#define DEF_Y	1

/* directions for create */
#define NW	0x01
#define NORTH	0x02
#define NE	0x04
#define EAST	0x08

/* directions for draw (also uses NORTH ans EAST) */
#define SOUTH	0x01
#define WEST	0x04
#define HBRIDGE	0x1F
#define VBRIDGE	0x2F

#define UD	0x01
#define LR	0x02
#define UL	0x04
#define UR	0x08
#define DL	0x10
#define DR	0x20

/* needed structures */
struct mazeloc {
  long rank;
  struct mazeloc *ptr;
};

struct nmazeloc {
  long x,y;
  int dirs;
};

struct hedgeloc {
  long x,y;
  int pos;
  int on;
};

/* global variables */
struct mazeloc *maze;
struct nmazeloc *nmaze;
struct nmazeloc **nmazelist;
struct hedgeloc *hedge;
struct hedgeloc **hedgelist;

/* number of real hedges */
long cur;

/* Determine number of bits. */
int bits[16] = {0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4};

/* all the options */
long bm;
int debug;
int exits;
int full;
int grey;
long paperh;
long lm;
int nobridge;
long pw;
int quiet;
long ph;
long scale;
long thick;
long vert;
long paperw;
long x;
long y;

void usage(void)
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,
    "bmaze [-q] [-d] [-n] [-e] [-f] [-x#] [-y#] [-s# [-v#]] [-p# [-r#]]\n");
  fprintf(stderr,"\t[-b#] [-l#] [-w#] [-h#] [-g#] [-t#]\n");
  exit(1);
}

void inits(void)
{
  long i,j,g;

  srand(time(NULL));

  /* malloc all the arrays for maze creation */

  if ((maze = (struct mazeloc *)malloc((size_t)
    (sizeof(struct mazeloc)*x*y)))==NULL) {
    if (!quiet) fprintf(stderr,"Could not malloc maze.\n");
    exit(1);
  }

  if ((hedge = (struct hedgeloc *)malloc((size_t)
    (sizeof(struct hedgeloc)*4*x*y)))==NULL) {
    if (!quiet) fprintf(stderr,"Could not malloc hedge.\n");
    exit(1);
  }

  if ((hedgelist = (struct hedgeloc **)malloc((size_t)
    (sizeof(struct hedgeloc *)*2*x*y)))==NULL) {
    if (!quiet) fprintf(stderr,"Could not malloc hedgelist.\n");
    exit(1);
  }

  /* init the maze */

  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      maze[x*j+i].rank = 0;
      maze[x*j+i].ptr = &maze[x*j+i];
    }
  }

  /* init the hedges */

  cur = 0;
  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      hedge[x*j+i].x = i;
      hedge[x*j+i].y = j;
      hedge[x*j+i].pos = NW;
      hedge[x*j+i].on = 0;
      if (!((i+j)&1) && i && (j<y-1)) hedgelist[cur++] = &hedge[x*j+i];
    }
  }

  g=x*y;
  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      hedge[g+x*j+i].x = i;
      hedge[g+x*j+i].y = j;
      hedge[g+x*j+i].pos = NORTH;
      hedge[g+x*j+i].on = 0;
      if (!((i+j)&1) && (j<y-2)) hedgelist[cur++] = &hedge[g+x*j+i];
    }
  }

  g += x*y;
  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      hedge[g+x*j+i].x = i;
      hedge[g+x*j+i].y = j;
      hedge[g+x*j+i].pos = NE;
      hedge[g+x*j+i].on = 0;
      if (!((i+j)&1) && (i<x-1) && (j<y-1)) hedgelist[cur++] = &hedge[g+x*j+i];
    }
  }

  g += x*y;
  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      hedge[g+x*j+i].x = i;
      hedge[g+x*j+i].y = j;
      hedge[g+x*j+i].pos = EAST;
      hedge[g+x*j+i].on = 0;
      if (!((i+j)&1) && (i<x-2)) hedgelist[cur++] = &hedge[g+x*j+i];
    }
  }

  if (debug) fprintf(stderr,"Maze initialised.\n");

}

struct mazeloc *find(struct mazeloc *a)
{
  return ( (a != a->ptr) ? find(a->ptr) : a->ptr );
}

void create(void)
{
  long i,j,k;
  struct mazeloc *a,*b;
  struct hedgeloc *htmp;

  /* Randomize the hedges. */

  for (i=cur-1;i;i--) {
    htmp = hedgelist[i];
    /* Make sure we have a long random number. */
    j = ((rand()&0x7FFF)<<16|(rand()&0xFF)<<8|(rand()&0xFF))%(i+1);
    hedgelist[i] = hedgelist[j];
    hedgelist[j] = htmp;
  }

  /* Now really create the maze. */

  for (k=0;k<cur;k++) {
    i = hedgelist[k]->x;
    j = hedgelist[k]->y;
    a = find(&maze[x*j+i]);
    switch (hedgelist[k]->pos) {
      case NW:
        b = find(&maze[x*(j+1)+i-1]);
        break;
      case NORTH:
        b = find(&maze[x*(j+2)+i]);
        break;
      case NE:
        b = find(&maze[x*(j+1)+i+1]);
        break;
      case EAST:
        b = find(&maze[x*j+i+2]);
        break;
    }
    if (a!=b && !(nobridge && ((hedgelist[k]->pos==NORTH && i &&
      hedge[3*x*y+x*(j+1)+i-1].on) || (hedgelist[k]->pos==EAST &&
      j && hedge[x*y+x*(j-1)+i+1].on)))) {
      if (a->rank == b->rank) {
	a->ptr = b;
	b->rank++;
      }
      else if (a->rank > b->rank) b->ptr = a;
      else a->ptr = b;
      hedgelist[k]->on = 1;
    }
  }

  /* Get rid of the useless memory. */

  free(maze);
  free(hedgelist);

}

int near(long i, long j)
{
  int result;

  result = 0;

  if (j && hedge[x*y+x*(j-1)+i].on) result |= UD;
  if (i && hedge[3*x*y+x*j+i-1].on) result |= LR;
  if (i && hedge[2*x*y+x*j+i-1].on) result |= UL;
  if (i<x-1 && hedge[x*j+i+1].on) result |= UR;
  if (j && hedge[x*(j-1)+i].on) result |= DL;
  if (j && hedge[2*x*y+x*(j-1)+i].on) result |= DR;

  return result;
}

void draw(void)
{
  long i,j,k,loc;
  struct nmazeloc *ntmp;
  int d,done;

  /* malloc nmaze just like hedges */

  if ((nmaze = (struct nmazeloc *)malloc((size_t)
    (sizeof(struct nmazeloc)*x*y)))==NULL) {
    if (!quiet) fprintf(stderr,"Could not malloc nmaze.\n");
    exit(1);
  }

  if ((nmazelist = (struct nmazeloc **)malloc((size_t)
    (sizeof(struct nmazeloc *)*x*y/2)))==NULL) {
    if (!quiet) fprintf(stderr,"Could not malloc nmazelist.\n");
    exit(1);
  }

  loc = 0;

  for(j=0;j<y;j++) {
    for(i=0;i<x;i++) {
      nmaze[x*j+i].dirs = 0;
      nmaze[x*j+i].x = i;
      nmaze[x*j+i].y = j;
      if ((i+j)&1) nmazelist[loc++] = &nmaze[x*j+i];
    }
  }

  /* randomize the nmaze locations */

  for (i=loc-1;i;i--) {
    ntmp = nmazelist[i];
    j = ((rand()&0x7FFF)<<16|(rand()&0xFF)<<8|(rand()&0xFF))%(i+1);
    nmazelist[i] = nmazelist[j];
    nmazelist[j] = ntmp;
  }

  cur = x*y-loc-1;

  if (debug) fprintf(stderr,"Maze created, %ld hedges.\n",cur);

  /* What follows is a tedious list of cases. As my editor is smarter than
     I am, this is the easiest way. */

  /* Step 1. The bridges. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UD|LR:
        nmazelist[k]->dirs = (rand()|1 ? HBRIDGE : VBRIDGE);
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[3*x*y+x*j+i-1].on = 0;
        cur -= 2;
        break;
      case UD|LR|UL:
        if (hedge[2*x*y+x*j+i-1].on == 1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
          hedge[2*x*y+x*j+i-1].on = 2;
          cur -= 2;
        }
        else if (rand()|1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST;
          hedge[2*x*y+x*j+i-1].on = 0;
          cur -= 3;
        }
        else {
          nmazelist[k]->dirs = SOUTH|EAST|WEST;
          hedge[2*x*y+x*j+i-1].on = 0;
          cur -= 3;
        }
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[3*x*y+x*j+i-1].on = 0;
        break;
      case UD|LR|UR:
        if (hedge[x*j+i+1].on == 1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
          hedge[x*j+i+1].on = 2;
          cur -= 2;
        }
        else if (rand()|1) {
          nmazelist[k]->dirs = NORTH|SOUTH|WEST;
          hedge[x*j+i+1].on = 0;
          cur -= 3;
        }
        else {
          nmazelist[k]->dirs = SOUTH|EAST|WEST;
          hedge[x*j+i+1].on = 0;
          cur -= 3;
        }
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[3*x*y+x*j+i-1].on = 0;
        break;
      case UD|LR|DL:
        if (hedge[x*(j-1)+i].on == 1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
          hedge[x*(j-1)+i].on = 2;
          cur -= 2;
        }
        else if (rand()|1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST;
          hedge[x*(j-1)+i].on = 0;
          cur -= 3;
        }
        else {
          nmazelist[k]->dirs = SOUTH|EAST|NORTH;
          hedge[x*(j-1)+i].on = 0;
          cur -= 3;
        }
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[3*x*y+x*j+i-1].on = 0;
        break;
      case UD|LR|DR:
        if (hedge[2*x*y+x*(j-1)+i].on == 1) {
          nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
          hedge[2*x*y+x*(j-1)+i].on = 2;
          cur -= 2;
        }
        else if (rand()|1) {
          nmazelist[k]->dirs = NORTH|SOUTH|WEST;
          hedge[2*x*y+x*(j-1)+i].on = 0;
          cur -= 3;
        }
        else {
          nmazelist[k]->dirs = NORTH|EAST|WEST;
          hedge[2*x*y+x*(j-1)+i].on = 0;
          cur -= 3;
        }
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[3*x*y+x*j+i-1].on = 0;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 1 done, %ld hedges.\n",cur);

  /* Step 1a. Remove the marked (on==2) hedges. */

  for (i=0;i<4*x*y;i++)
    if (hedge[i].on==2) {
      hedge[i].on = 0;
      cur--;
    }

  if (debug) fprintf(stderr,"Draw step 1a done, %ld hedges.\n",cur);

  /* Step 2. Single long and 2 short hedges. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UD|UL|UR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        cur -= 3;
        break;
      case UD|UL|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case UD|DL|UR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[x*(j-1)+i].on = 0;
        hedge[x*j+i+1].on = 0;
        cur -= 3;
        break;
      case UD|DL|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[x*(j-1)+i].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case LR|UL|DL:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case LR|UL|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case LR|UR|DL:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case LR|UR|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 2 done, %ld hedges.\n",cur);

  /* Step 3. Single long and single short hedge. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UD|UL:
        nmazelist[k]->dirs = NORTH|SOUTH|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        cur -= 2;
        break;
      case UD|UR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[x*j+i+1].on = 0;
        cur -= 2;
        break;
      case UD|DL:
        nmazelist[k]->dirs = NORTH|SOUTH|WEST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 2;
        break;
      case UD|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST;
        hedge[x*y+x*(j-1)+i].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 2;
        break;
      case LR|UL:
        nmazelist[k]->dirs = WEST|EAST|NORTH;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*j+i-1].on = 0;
        cur -= 2;
        break;
      case LR|UR:
        nmazelist[k]->dirs = WEST|EAST|NORTH;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        cur -= 2;
        break;
      case LR|DL:
        nmazelist[k]->dirs = WEST|EAST|SOUTH;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 2;
        break;
      case LR|DR:
        nmazelist[k]->dirs = WEST|EAST|SOUTH;
        hedge[3*x*y+x*j+i-1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 2;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 3 done, %ld hedges.\n",cur);

  /* Step 4. A single long hedge. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UD:
        nmazelist[k]->dirs = NORTH|SOUTH;
        hedge[x*y+x*(j-1)+i].on = 0;
        cur--;
        break;
      case LR:
        nmazelist[k]->dirs = WEST|EAST;
        hedge[3*x*y+x*j+i-1].on = 0;
        cur--;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 4 done, %ld hedges.\n",cur);

  /* Step 5. Three short hedges. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UL|UR|DL:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case UL|UR|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case UL|DL|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
      case UR|DL|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST|WEST;
        hedge[x*j+i+1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 3;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 5 done, %ld hedges.\n",cur);

  /* Step 6. Two adjacent short hedges. */

  for (k=0;k<loc;k++) {
    i = nmazelist[k]->x;
    j = nmazelist[k]->y;
    switch (near(i,j)) {
      case UL|UR:
        nmazelist[k]->dirs = NORTH|WEST|EAST;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*j+i+1].on = 0;
        cur -= 2;
        break;
      case UL|DL:
        nmazelist[k]->dirs = NORTH|SOUTH|WEST;
        hedge[2*x*y+x*j+i-1].on = 0;
        hedge[x*(j-1)+i].on = 0;
        cur -= 2;
        break;
      case UR|DR:
        nmazelist[k]->dirs = NORTH|SOUTH|EAST;
        hedge[x*j+i+1].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 2;
        break;
      case DL|DR:
        nmazelist[k]->dirs = SOUTH|EAST|WEST;
        hedge[x*(j-1)+i].on = 0;
        hedge[2*x*y+x*(j-1)+i].on = 0;
        cur -= 2;
        break;
    }
  }

  if (debug) fprintf(stderr,"Draw step 6 done, %ld hedges.\n",cur);

  /* Step 7. A single short hedge. */

  while (cur) {
    for (k=0;k<loc;k++) {
      i = nmazelist[k]->x;
      j = nmazelist[k]->y;
      switch (near(i,j)) {
        case UL:
          nmazelist[k]->dirs = NORTH|WEST;
          hedge[2*x*y+x*j+i-1].on = 0;
          cur--;
          break;
        case UR:
          nmazelist[k]->dirs = NORTH|EAST;
          hedge[x*j+i+1].on = 0;
          cur--;
          break;
        case DL:
          nmazelist[k]->dirs = SOUTH|WEST;
          hedge[x*(j-1)+i].on = 0;
          cur--;
          break;
        case DR:
          nmazelist[k]->dirs = SOUTH|EAST;
          hedge[2*x*y+x*(j-1)+i].on = 0;
          cur--;
          break;
      }
    }
    if (debug) fprintf(stderr,"Draw step 7 done, %ld hedges.\n",cur);
  }

  free(hedge);

  /* Now make the maze locations conform to the nmaze ones. */

  for (j=0;j<y;j++) {
    for (i=0;i<x;i++) {
      if (!((i+j)&1)) {
        if (j && nmaze[x*(j-1)+i].dirs & NORTH)
          nmaze[x*j+i].dirs |= SOUTH;
        if (j<y-1 && nmaze[x*(j+1)+i].dirs & SOUTH)
          nmaze[x*j+i].dirs |= NORTH;
        if (i && nmaze[x*j+i-1].dirs & EAST)
          nmaze[x*j+i].dirs |= WEST;
        if (i<x-1 && nmaze[x*j+i+1].dirs & WEST)
          nmaze[x*j+i].dirs |= EAST;
      }
    }
  }

  /* Fill the empty squares. */
  for (k=0;k<loc;k++) {
    if (!nmazelist[k]->dirs) {
      i = nmazelist[k]->x;
      j = nmazelist[k]->y;
      d = 0;
      done = 0;
      while (!done && d!=15) {
        switch ((int)(rand()&3)) {
          case 0:
            if (i && bits[nmaze[x*j+i-1].dirs]==1) {
              nmaze[x*j+i].dirs |= WEST;
              nmaze[x*j+i-1].dirs |= EAST;
              done = 1;
            }
            else d |= WEST;
            break;
          case 1:
            if (i<x-1 && bits[nmaze[x*j+i+1].dirs]==1) {
              nmaze[x*j+i].dirs |= EAST;
              nmaze[x*j+i+1].dirs |= WEST;
              done = 1;
            }
            else d |= EAST;
            break;
          case 2:
            if (j && bits[nmaze[x*(j-1)+i].dirs]==1) {
              nmaze[x*j+i].dirs |= SOUTH;
              nmaze[x*(j-1)+i].dirs |= NORTH;
              done = 1;
            }
            else d |= SOUTH;
            break;
          case 3:
            if (j<y-1 && bits[nmaze[x*(j+1)+i].dirs]==1) {
              nmaze[x*j+i].dirs |= NORTH;
              nmaze[x*(j+1)+i].dirs |= SOUTH;
              done = 1;
            }
            else d |= NORTH;
            break;
        }
      }
      while (!done) {
        switch ((int)(rand()&3)) {
          case 0:
            if (i) {
              nmaze[x*j+i].dirs |= WEST;
              nmaze[x*j+i-1].dirs |= EAST;
              done = 1;
            }
            break;
          case 1:
            if (i<x-1) {
              nmaze[x*j+i].dirs |= EAST;
              nmaze[x*j+i+1].dirs |= WEST;
              done = 1;
            }
            break;
          case 2:
            if (j) {
              nmaze[x*j+i].dirs |= SOUTH;
              nmaze[x*(j-1)+i].dirs |= NORTH;
              done = 1;
            }
            break;
          case 3:
            if (j<y-1) {
              nmaze[x*j+i].dirs |= NORTH;
              nmaze[x*(j+1)+i].dirs |= SOUTH;
              done = 1;
            }
            break;
        }
      }
    }
  }

  if (debug) fprintf(stderr,"Empty squares filled.\n");

  free(nmazelist);

  /* Make entrance and exit. */

  if (!exits) {
    /* This is temporary ! */
    nmaze[rand()%x].dirs |= SOUTH;
    nmaze[x*(y-1)+rand()%x].dirs |= NORTH;
  }

}

void print(void)
{
  long xpages,ypages,xp,yp;
  int d;
  long xst,yst,xnum,ynum;
  long xw,yw,xnl,ynl;
  long i,j;
  long hw,vw,hm,vm;

  xw = paperw/scale;
  yw = paperh/vert;
  xpages = (x+xw-1)/xw;
  ypages = (y+yw-1)/yw;
  if ((xnl = x%xw)==0) xnl=xw;
  if ((ynl = y%yw)==0) ynl=yw;

  hm = scale*(100-pw)/200;
  hw = scale-2*hm;
  vm = vert*(100-ph)/200;
  vw = vert-2*vm;

  /* First some conforming PS blurb */

  printf("%%!PS-Adobe-1.0\n");
  printf("%%%%Creator: bmaze by Bas de Bakker\n");
  printf("%%%%Title: Mazes unlimited\n");
  printf("%%%%CreationDate: Big Bang\n");
  printf("%%%%Pages: %ld\n", xpages * ypages);
  printf("%%%%DocumentFonts:\n");
  printf("%%%%BoundingBox: %ld %ld %ld %ld\n",lm,bm,
    lm+(xpages==1?x*scale:xw*scale),bm+(ypages==1?y*vert:yw*vert));
  printf("%%%%EndComments\n");
  printf("/M { moveto } def\n/L { lineto } def\n");
  printf("/T { %ld 0 translate } def\n",scale);
  printf("/H { %ld %ld M %ld %ld L ",hm,vm,scale+hm+hw,vm);
  printf("%ld %ld L %ld %ld L closepath fill } def\n",
    scale+hm+hw,vm+vw,hm,vm+vw);
  printf("/V { %ld %ld M %ld %ld L ",hm,vm,hm+hw,vm);
  printf("%ld %ld L %ld %ld L closepath fill } def\n",
    hm+hw,vert+vm+vw,hm,vert+vm+vw);
  printf("/ST { stroke %ld 0 translate } def\n",scale);
  printf("/N { %ld %ld M %ld %ld L ",hm,vm+vw,hm,vert);
  printf("%ld %ld M %ld %ld L } def\n",hm+hw,vm+vw,hm+hw,vert);
  printf("/S { %ld 0 M %ld %ld L ",hm,hm,vm);
  printf("%ld 0 M %ld %ld L } def\n",hm+hw,hm+hw,vm);
  printf("/E { %ld %ld M %ld %ld L ",hm+hw,vm,scale,vm);
  printf("%ld %ld M %ld %ld L } def\n",hm+hw,vm+vw,scale,vm+vw);
  printf("/W { 0 %ld M %ld %ld L ",vm,hm,vm);
  printf("0 %ld M %ld %ld L } def\n",vm+vw,hm,vm+vw);
  printf("/NB { %ld %ld M %ld %ld L } def\n",hm,vm+vw,hm+hw,vm+vw);
  printf("/SB { %ld %ld M %ld %ld L } def\n",hm,vm,hm+hw,vm);
  printf("/EB { %ld %ld M %ld %ld L } def\n",hm+hw,vm,hm+hw,vm+vw);
  printf("/WB { %ld %ld M %ld %ld L } def\n",hm,vm,hm,vm+vw);
  printf("%%%%EndProlog\n");

  /* Use nmaze[].x and y for used markers. */

  for (i=0;i<x*y;i++) {
    nmaze[i].x = 0;
    nmaze[i].y = 0;
  }

  for (yp=0;yp<ypages;yp++) {
    for (xp=0;xp<xpages;xp++) {

      printf("%%%%Page: %ld,%ld %ld\n", xp, yp, xpages * yp + xp + 1);
      printf("save\n%ld %ld translate\n",lm,bm);

      xst = (long)xp*xw;
      yst = (long)yp*yw;
      xnum = xp==xpages-1 ? xnl : xw;
      ynum = yp==ypages-1 ? ynl : yw;

      printf("newpath\n");
      printf("0 0 M\n%ld 0 L\n%ld %ld L\n0 %ld L\n",
        scale*xnum,scale*xnum,vert*ynum,vert*ynum);
      printf("closepath\nclip\n");

      /* First do the background (of course). */

      printf("%4.2f setgray\nfill\n",(float)grey/100);

      /* Now the paths themselves. */

      printf("1 setgray\ngsave\n");
      printf("0 %ld translate\n",-vert);
      for (i=0;i<xnum;i++) {
        if (nmaze[x*yst+xst+i].dirs&SOUTH) printf("V ");
        printf("T\n");
      }
      for (j=0;j<ynum;j++) {
        printf("%ld %ld translate\n",-scale*(xnum+1),vert);
        if (nmaze[x*(yst+j)+xst].dirs&WEST) printf("H ");
        printf("T\n");
        for (i=0;i<xnum;i++) {
          if (nmaze[x*(yst+j)+xst+i].dirs&EAST)
            printf("H ");
          if (nmaze[x*(yst+j)+xst+i].dirs&NORTH)
            printf("V ");
          printf("T\n");
        }
      }
      printf("grestore\n");

      /* Now the lines around the paths. */

      printf("0 setgray\n2 setlinecap\n");
      printf("%4.1f setlinewidth\n",(float)thick/10);

      for (j=0;j<ynum;j++) {
        for (i=0;i<xnum;i++) {
          d = nmaze[x*(yst+j)+xst+i].dirs;
          if (d&NORTH) printf("N ");
          if (d&SOUTH) printf("S ");
          if (d&EAST) printf("E ");
          if (d&WEST) printf("W ");
          if (!(d&NORTH) || d==HBRIDGE) printf("NB ");
          if (!(d&SOUTH) || d==HBRIDGE) printf("SB ");
          if (!(d&EAST) || d==VBRIDGE) printf("EB ");
          if (!(d&WEST) || d==VBRIDGE) printf("WB ");
          printf("ST\n");
        }
        printf("%ld %ld translate\n", -scale * xnum, vert);
      }

      printf("showpage\nrestore\n");

    }
  }

  printf("%%%%Trailer\n");

  /* That's all folks ! */

  free(nmaze);
}

int main(int argc, char **argv)
{
  int i;

  bm = DEF_B;
  debug = 0;
  exits = 0;
  full = 0;
  grey = DEF_G;
  paperh = DEF_H;
  lm = DEF_L;
  nobridge = 0;
  pw = DEF_P;
  quiet = 0;
  ph = DEF_R;
  scale = DEF_S;
  thick = DEF_T;
  vert = DEF_V;
  paperw = DEF_W;
  x = DEF_X;
  y = DEF_Y;

  for (i=1;i<argc;i++) {
    if (*argv[i] != '-') usage();
    argv[i]++;
    switch (*argv[i]++) {
      case 'b':
        if ((bm = atol(argv[i]))==0) usage();
        break;
      case 'd':
        debug = 1;
        break;
      case 'e':
	exits = 1;
	break;
      case 'f':
	full = 1;
	break;
      case 'g':
        if ((grey = atoi(argv[i]))==0) usage();
        break;
      case 'h':
        if ((paperh = atol(argv[i]))==0) usage();
        break;
      case 'l':
        if ((lm = atol(argv[i]))==0) usage();
        break;
      case 'n':
        nobridge = 1;
        break;
      case 'p':
        if ((pw = atol(argv[i]))==0) usage();
        break;
      case 'q':
        quiet = 1;
        break;
      case 'r':
        if ((ph = atol(argv[i]))==0) usage();
        break;
      case 's':
        if ((scale = atol(argv[i]))==0) usage();
        break;
      case 't':
        if ((thick = atol(argv[i]))==0) usage();
        break;
      case 'v':
        if ((vert = atol(argv[i]))==0) usage();
        break;
      case 'w':
        if ((paperw = atol(argv[i]))==0) usage();
        break;
      case 'x':
        if ((x = atol(argv[i]))==0) usage();
        break;
      case 'y':
        if ((y = atol(argv[i]))==0) usage();
        break;
      default:
        usage();
    }
  }

  if (vert == 0) vert = scale ;
  if (ph == 0) ph = pw;
  if (!full) {
    x *= paperw/scale;
    y *= paperh/vert;
  }

  inits();
  create();
  draw();
  print();
  return 0;
}
